from jnpr.junos.device import Device
from . import jxml
from . import jxml as JXML
from . import version
from . import exception

__version__ = version.VERSION
__date__ = version.DATE

#import time
#__date__ = time.strftime("%Y-%b-%d")
